use ::reader::Reader;
use ::lexer::token::Token;
use ::lexer::token_types::TokenTypes;



/// The keywords for this language.
const keywords: [&'static str; 1] = ["include"];



/// Performs lexigraphical analysis on
/// the text given from a Reader.
pub enum Lexer
{
}



/// Determines if a character is alphabetic.
fn is_alpha(c: char) -> bool
{
   c.is_alphabetic()
}

/// Determines if a character is a numeral
fn is_numeric(c: char) -> bool
{
   c.is_numeric()
}

/// Determines if a character is a numeral
fn is_alphanumeric(c: char) -> bool
{
   is_alpha(c) | is_numeric(c)
}

/// Determines if a character is whitespace.
fn is_whitespace(c: char) -> bool
{
   c.is_whitespace()
}

fn is_non_whitespace(c: char) -> bool
{
   !is_whitespace(c)
}

/// Determines if a character is a symbol.
fn is_symbol(c: char) -> bool
{
   is_whitespace(c) == false && is_alphanumeric(c) == false
}

/// Determines if a String is a keyword.
fn is_keyword(s: String) -> bool
{
   for keyword in keywords.iter()
   {
      if &s == keyword
      {
         return true
      }
   }

   false
}

/// Returns true if the next set of characters starts with
/// the given pattern; Otherwise, false.
fn starts_with<P>(pattern: P) -> bool
   where P: AsRef<str>
{
   false
}

/// Consume and discard zero or more whitespace characters.
fn consume_whitespace(reader: &mut Reader) -> String
{
   consume_while(reader, is_whitespace)
}

///
fn consume_symbol(reader: &mut Reader) -> String
{
   let mut is_valid: bool;
   let mut buffer: String;

   // Create a buffer to store all the consumed characters.
   buffer = String::new();

   // Check if the first character from the Reader
   // is a symbol character.
   match reader.get_char()
   {
      Ok(read_character) =>
      {
         is_valid = is_symbol(read_character);
      }

      Err(error) =>
      {
         warn!("{}", error);
         is_valid = false;
      }
   }

   if is_valid == true
   {
      // Consume this character that passed.
      match reader.consume_char()
      {
         Ok(character) =>
         {
            // Add this consumed character to the buffer.
            buffer.push(character);
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }

   // Return the buffer of consumed characters.
   buffer
}

///
fn consume_numeral(reader: &mut Reader) -> String
{
   let mut is_valid: bool;
   let mut buffer: String;

   // Create a buffer to store all the consumed characters.
   buffer = String::new();

   // Check if the first character from the Reader
   // is a numeric character.
   match reader.get_char()
   {
      Ok(read_character) =>
      {
         is_valid = is_numeric(read_character);
      }

      Err(error) =>
      {
         warn!("{}", error);
         is_valid = false;
      }
   }

   // If the first character is valid, then consume it and
   // consume any numeric characters that follow.
   if is_valid == true
   {
      // Consume this character that passed.
      match reader.consume_char()
      {
         Ok(character) =>
         {
            // Add this consumed character to the buffer.
            buffer.push(character);
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }

      // Now read any numeric characters that follow.
      buffer.push_str(&consume_while(reader, is_non_whitespace));
   }

   // Return the buffer of consumed characters.
   buffer
}

///
fn consume_identifier(reader: &mut Reader) -> String
{
   let mut is_valid: bool;
   let mut buffer: String;

   // Create a buffer to store all the consumed characters.
   buffer = String::new();

   // Check if the first character from the Reader
   // is an alpha character.
   match reader.get_char()
   {
      Ok(read_character) =>
      {
         is_valid = is_alpha(read_character);
      }

      Err(error) =>
      {
         warn!("{}", error);
         is_valid = false;
      }
   }

   // If the first character is valid, then consume it and
   // consume any alphanumeric character the follow.
   if is_valid == true
   {
      // Consume this character that passed.
      match reader.consume_char()
      {
         Ok(character) =>
         {
            // Add this consumed character to the buffer.
            buffer.push(character);
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }

      // Now read any alphanumeric characters that follow.
      buffer.push_str(&consume_while(reader, is_alphanumeric));
   }

   // Return the buffer of consumed characters.
   buffer
}

/// Consume characters until the test returns false.
fn consume_while<F>(reader: &mut Reader, test: F) -> String
   where F: Fn(char) -> bool
{
   let mut is_valid: bool;
   let mut buffer: String;

   // Create a buffer to store all the consumed characters.
   buffer = String::new();

   // Get a character from the reader.
   match reader.get_char()
   {
      Ok(read_character) =>
      {
         is_valid = test(read_character);
      }

      Err(error) =>
      {
         warn!("{}", error);
         is_valid = false;
      }
   }

   // Loop until we get to invalid input.
   while is_valid == true
   {
      // Consume this character that passed.
      match reader.consume_char()
      {
         Ok(character) =>
         {
            // Add this consumed character to the buffer.
            buffer.push(character);

            // Get a character from the reader.
            match reader.get_char()
            {
               Ok(read_character) =>
               {
                  is_valid = test(read_character);
               }

               Err(error) =>
               {
                  warn!("{}", error);
                  is_valid = false;
               }
            }
         }

         Err(error) =>
         {
            is_valid = false;
            error!("{}", error);
         }
      }
   }

   // Return the buffer of consumed characters.
   buffer
}



impl Lexer
{
   /// Scan the data held by the Reader. This will consume the Reader and
   /// clear it.
   pub fn scan(reader: &mut Reader) -> Result<Vec<Token>, &'static str>
   {
      let mut test_char: char;
      let mut token: Token;
      let mut tokens: Vec<Token>;

      // Create a new list to hold all the discovered tokens.
      tokens = Vec::new();

      // Begin turning the input into tokens.
      while reader.is_eob() == false
      {
         test_char = try!(reader.get_char());
         if is_whitespace(test_char)
         {
            // Just skip/remove any whitespace.
            token = Token::from(consume_whitespace(reader));
            token.set_type(TokenTypes::Whitespace);

            // Push the token into the list.
            tokens.push(token);
         }
         else if is_numeric(test_char)
         {
            // The next character is a numeric,
            // so we are probably looking at a numeral.
            token = Token::from(consume_numeral(reader));
            token.set_type(TokenTypes::Numeric);

            // Push the token into the list.
            tokens.push(token);
         }
         else if is_alpha(test_char)
         {
            // The next character is an alpha character,
            // so we are probably looking at an identifier.
            token = Token::from(consume_identifier(reader));

            // Determine if this identifier is a keyword.
            if is_keyword(token.to_string()) == true
            {
               token.set_type(TokenTypes::Keyword);
            }
            else
            {
               token.set_type(TokenTypes::Identifier);
            }

            // Push the token into the list.
            tokens.push(token);
         }
         else
         {
            // This must be some kind of symbol.
            token = Token::from(consume_symbol(reader));
            token.set_type(TokenTypes::Unknown);

            // Push the token into the list.
            tokens.push(token);
         }
      }

      // Clear the reader since we are finished with it.
      reader.clear();

for token in tokens.iter()
{
   println!("{}: {}", token.get_type(), token);
}

      // Return the scanned tokens.
      Ok(tokens)
   }
}
